import os
import string
from pathlib import Path
from typing import Optional

from .types import TimeAlignedType


def _get_duration(name: str) -> float:
    import soundfile

    from sonusai import SonusAIError

    try:
        return soundfile.info(name).duration
    except Exception as e:
        raise SonusAIError(f'Error reading {name}: {e}')


def load_text(audio: str | os.PathLike[str]) -> Optional[TimeAlignedType]:
    """Load time-aligned text data given a L2-ARCTIC audio file.

    :param audio: Path to the L2-ARCTIC audio file.
    :return: A TimeAlignedType object.
    """
    file = Path(audio).parent.parent / 'transcript' / (Path(audio).stem + '.txt')
    if not os.path.exists(file):
        return None

    with open(file, mode='r', encoding='utf-8') as f:
        line = f.read()

    return TimeAlignedType(0,
                           _get_duration(str(audio)),
                           line.strip().lower().translate(str.maketrans('', '', string.punctuation)))


def load_words(audio: str | os.PathLike[str]) -> Optional[list[TimeAlignedType]]:
    """Load time-aligned word data given a L2-ARCTIC audio file.

    :param audio: Path to the L2-ARCTIC audio file.
    :return: A list of TimeAlignedType objects.
    """
    return _load_ta(audio, 'words')


def load_phonemes(audio: str | os.PathLike[str]) -> Optional[list[TimeAlignedType]]:
    """Load time-aligned phonemes data given a L2-ARCTIC audio file.

    :param audio: Path to the L2-ARCTIC audio file.
    :return: A list of TimeAlignedType objects.
    """
    return _load_ta(audio, 'phones')


def _load_ta(audio: str | os.PathLike[str], tier: str) -> Optional[list[TimeAlignedType]]:
    from praatio import textgrid

    file = Path(audio).parent.parent / 'textgrid' / (Path(audio).stem + '.TextGrid')
    if not os.path.exists(file):
        return None

    tg = textgrid.openTextgrid(str(file), includeEmptyIntervals=False)
    if tier not in tg.tierNames:
        return None

    entries: list[TimeAlignedType] = []
    for entry in tg.getTier(tier).entries:
        entries.append(TimeAlignedType(text=entry.label, start=entry.start, end=entry.end))

    return entries


def load_annotations(audio: str | os.PathLike[str]) -> Optional[dict[str, list[TimeAlignedType]]]:
    """Load time-aligned annotation data given a L2-ARCTIC audio file.

    :param audio: Path to the L2-ARCTIC audio file.
    :return: A dictionary of a list of TimeAlignedType objects.
    """
    from praatio import textgrid

    file = Path(audio).parent.parent / 'annotation' / (Path(audio).stem + '.TextGrid')
    if not os.path.exists(file):
        return None

    tg = textgrid.openTextgrid(str(file), includeEmptyIntervals=False)
    result: dict[str, list[TimeAlignedType]] = {}
    for tier in tg.tierNames:
        entries: list[TimeAlignedType] = []
        for entry in tg.getTier(tier).entries:
            entries.append(TimeAlignedType(text=entry.label, start=entry.start, end=entry.end))
        result[tier] = entries

    return result


def load_speakers(input_dir: Path) -> dict:
    speakers = {}
    with open(input_dir / 'readme-download.txt') as file:
        processing = False
        for line in file:
            if not processing and line.startswith('|---|'):
                processing = True
                continue

            if processing:
                if line.startswith('|**Total**|'):
                    break
                else:
                    fields = line.strip().split('|')
                    speaker_id = fields[1]
                    gender = fields[2]
                    dialect = fields[3]
                    speakers[speaker_id] = {'gender': gender, 'dialect': dialect}

    return speakers



"""sonusai genmetrics

usage: genmetrics [-hvs] [-i MIXID] [-n INCLUDE] [-x EXCLUDE] LOC

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -i MIXID, --mixid MIXID         Mixture ID(s) to generate. [default: *].
    -n INCLUDE, --include INCLUDE   Metrics to include. [default: all]
    -x EXCLUDE, --exclude EXCLUDE   Metrics to exclude. [default: none]
    -s, --supported                 Show list of supported metrics.

Calculate speech enhancement metrics of SonusAI mixture data in LOC.

Inputs:
    LOC         A SonusAI mixture database directory.
    MIXID       A glob of mixture ID(s) to generate.
    INCLUDE     Comma separated list of metrics to include. Can be 'all' or
                any of the supported metrics.
    EXCLUDE     Comma separated list of metrics to exclude. Can be 'none' or
                any of the supported metrics.

Examples:

Generate all available mxwer metrics (as determined by mixdb asr_configs parameter):
> sonusai genmetrics -n"mxwer" mixdb_loc

Generate only mxwer.faster metrics:
> sonusai genmetrics -n"mxwer.faster" mixdb_loc

Generate all available metrics except for mxwer.faster:
> sonusai genmetrics -x"mxwer.faster" mixdb_loc

"""
import signal
from dataclasses import dataclass

from sonusai.mixture import MixtureDatabase


def signal_handler(_sig, _frame):
    import sys

    from sonusai import logger

    logger.info('Canceled due to keyboard interrupt')
    sys.exit(1)


signal.signal(signal.SIGINT, signal_handler)


@dataclass
class MPGlobal:
    mixdb: MixtureDatabase = None
    metrics: set[str] = None


MP_GLOBAL = MPGlobal()


def _initializer(location: str, metrics: set[str]) -> None:
    MP_GLOBAL.mixdb = MixtureDatabase(location)
    MP_GLOBAL.metrics = metrics


def _process_mixture(mixid: int) -> None:
    from sonusai.mixture import write_mixture_data

    mixdb = MP_GLOBAL.mixdb
    metrics = list(MP_GLOBAL.metrics)

    values = mixdb.mixture_metrics(m_id=mixid, metrics=metrics, force=True)
    write_data = list(zip(metrics, values))

    write_mixture_data(mixdb, mixdb.mixture(mixid), write_data)


def main() -> None:
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    mixids = args['--mixid']
    includes = [x.strip() for x in args['--include'].lower().split(',')]
    excludes = [x.strip() for x in args['--exclude'].lower().split(',')]
    show_supported = args['--supported']
    location = args['LOC']

    import sys
    import time
    from os.path import join

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler
    from sonusai.utils import pp_tqdm_imap
    from sonusai.utils import seconds_to_hms
    from tqdm import tqdm

    start_time = time.monotonic()

    # Setup logging file
    create_file_handler(join(location, 'genmetrics.log'))
    update_console_handler(verbose)
    initial_log_messages('genmetrics')

    logger.info(f'Load mixture database from {location}')

    mixdb = MixtureDatabase(location)
    supported = mixdb.supported_metrics
    if show_supported:
        logger.info(f'\nSupported metrics: {", ".join(sorted(supported))}')
        sys.exit(0)

    if includes is None or 'all' in includes:
        metrics = supported
    else:
        metrics = set(includes)
        if 'mxwer' in metrics:
            metrics.remove('mxwer')
            for name in mixdb.asr_configs:
                metrics.add(f'mxwer.{name}')

    diff = metrics.difference(supported)
    if diff:
        logger.error(f'Unrecognized metric: {", ".join(diff)}')
        sys.exit(1)

    if excludes is None or 'none' in excludes:
        _excludes = set([])
    else:
        _excludes = set(excludes)
        if 'mxwer' in _excludes:
            _excludes.remove('mxwer')
            for name in mixdb.asr_configs:
                _excludes.add(f'mxwer.{name}')

    diff = _excludes.difference(supported)
    if diff:
        logger.error(f'Unrecognized metric: {", ".join(diff)}')
        sys.exit(1)

    for exclude in _excludes:
        metrics.discard(exclude)

    logger.info(f'Generating metrics: {", ".join(metrics)}')

    mixids = mixdb.mixids_to_list(mixids)
    logger.info('')
    logger.info(f'Found {len(mixids):,} mixtures to process')

    progress = tqdm(total=len(mixids), desc='genmetrics')
    pp_tqdm_imap(_process_mixture, mixids,
                 progress=progress,
                 initializer=_initializer,
                 initargs=(location, metrics))
    progress.close()

    logger.info(f'Wrote metrics for {len(mixids)} mixtures to {location}')
    logger.info('')

    end_time = time.monotonic()
    logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')
    logger.info('')


if __name__ == '__main__':
    main()



from collections import namedtuple

TruthSettingRecord = namedtuple('TruthSettingRecord', [
    'id',
    'setting'])

TargetFileRecord = namedtuple('TargetFileRecord', [
    'id',
    'name',
    'samples',
    'level_type',
    'speaker_id'])

NoiseFileRecord = namedtuple('NoiseFileRecord', [
    'id',
    'name',
    'samples'])

TopRecord = namedtuple('TopRecord', [
    'id',
    'version',
    'class_balancing',
    'feature',
    'noise_mix_mode',
    'num_classes',
    'seed',
    'truth_mutex',
    'truth_reduction_function',
    'mixid_width',
    'speaker_metadata_tiers',
    'textgrid_metadata_tiers'])

ClassLabelRecord = namedtuple('ClassLabelRecord', [
    'id',
    'label'])

ClassWeightsThresholdRecord = namedtuple('ClassWeightsThresholdRecord', [
    'id',
    'threshold'])

ImpulseResponseFileRecord = namedtuple('ImpulseResponseFileRecord', [
    'id',
    'file'])

SpectralMaskRecord = namedtuple('SpectralMaskRecord', [
    'id',
    'f_max_width',
    'f_num',
    't_max_width',
    't_num',
    't_max_percent'])

TargetRecord = namedtuple('TargetRecord', [
    'id',
    'file_id',
    'augmentation',
    'gain'])

MixtureRecord = namedtuple('MixtureRecord', [
    'id',
    'name',
    'noise_file_id',
    'noise_augmentation',
    'noise_offset',
    'noise_snr_gain',
    'random_snr',
    'snr',
    'samples',
    'spectral_mask_id',
    'spectral_mask_seed',
    'target_snr_gain'
])



import os
from pathlib import Path
from typing import Optional

from .types import TimeAlignedType


def _get_duration(name: str) -> float:
    import soundfile

    from sonusai import SonusAIError

    try:
        return soundfile.info(name).duration
    except Exception as e:
        raise SonusAIError(f'Error reading {name}: {e}')


def load_text(audio: str | os.PathLike[str]) -> Optional[TimeAlignedType]:
    """Load time-aligned text data given a VCTK audio file.

    :param audio: Path to the VCTK audio file.
    :return: A TimeAlignedType object.
    """
    import string

    file = Path(audio).parents[2] / 'txt' / Path(audio).parent.name / (Path(audio).stem[:-5] + '.txt')
    if not os.path.exists(file):
        return None

    with open(file, mode='r', encoding='utf-8') as f:
        line = f.read()

    start = 0
    end = _get_duration(str(audio))
    text = line.strip().lower().translate(str.maketrans('', '', string.punctuation))

    return TimeAlignedType(start, end, text)


def load_speakers(input_dir: Path) -> dict:
    speakers = {}
    with open(input_dir / 'speaker-info.txt') as file:
        for line in file:
            if not line.startswith('ID'):
                fields = line.strip().split('(', 1)[0].split()
                speaker_id = fields[0]
                age = fields[1]
                gender = fields[2]
                dialect = ' '.join([field for field in fields[3:]])
                speakers[speaker_id] = {'gender': gender, 'dialect': dialect, 'age': age}
    return speakers



from typing import Optional

from sonusai.mixture.datatypes import AudioF
from sonusai.mixture.datatypes import AudioT
from sonusai.mixture.datatypes import Feature


def get_feature_from_audio(audio: AudioT,
                           feature_mode: str,
                           num_classes: Optional[int] = 1,
                           truth_mutex: Optional[bool] = False) -> Feature:
    """Apply forward transform and generate feature data from audio data

    :param audio: Time domain audio data [samples]
    :param feature_mode: Feature mode
    :param num_classes: Number of classes
    :param truth_mutex: Whether to calculate 'other' label
    :return: Feature data [frames, strides, feature_parameters]
    """
    import numpy as np
    from pyaaware import FeatureGenerator

    from .augmentation import pad_audio_to_frame
    from .datatypes import TransformConfig
    from .helpers import forward_transform

    fg = FeatureGenerator(feature_mode=feature_mode,
                          num_classes=num_classes,
                          truth_mutex=truth_mutex)

    feature_step_samples = fg.ftransform_R * fg.decimation * fg.step
    audio = pad_audio_to_frame(audio, feature_step_samples)

    audio_f = forward_transform(audio=audio,
                                config=TransformConfig(N=fg.ftransform_N,
                                                       R=fg.ftransform_R,
                                                       bin_start=fg.bin_start,
                                                       bin_end=fg.bin_end,
                                                       ttype=fg.ftransform_ttype))

    samples = len(audio)
    transform_frames = samples // fg.ftransform_R
    feature_frames = samples // feature_step_samples

    feature = np.empty((feature_frames, fg.stride, fg.feature_parameters), dtype=np.float32)

    feature_frame = 0
    for transform_frame in range(transform_frames):
        fg.execute(audio_f[transform_frame])

        if fg.eof():
            feature[feature_frame] = fg.feature()
            feature_frame += 1

    return feature


def get_audio_from_feature(feature: Feature,
                           feature_mode: str,
                           num_classes: Optional[int] = 1,
                           truth_mutex: Optional[bool] = False,
                           trim: Optional[bool] = True) -> AudioT:
    """Apply inverse transform to feature data to generate audio data

    :param feature: Feature data [frames, strides, feature_parameters]
    :param feature_mode: Feature mode
    :param num_classes: Number of classes
    :param truth_mutex: Whether to calculate 'other' label
    :param trim: Whether to trim the audio data
    :return: Audio data [samples]
    """
    import numpy as np

    from pyaaware import FeatureGenerator

    from .datatypes import TransformConfig
    from .helpers import inverse_transform
    from sonusai.utils.stacked_complex import unstack_complex

    fg = FeatureGenerator(feature_mode=feature_mode,
                          num_classes=num_classes,
                          truth_mutex=truth_mutex)

    feature_complex = unstack_complex(feature)
    if feature_mode[0:1] == 'h':
        feature_complex = _power_uncompress(feature_complex)
    return np.squeeze(inverse_transform(transform=feature_complex,
                                        config=TransformConfig(N=fg.itransform_N,
                                                               R=fg.itransform_R,
                                                               bin_start=fg.bin_start,
                                                               bin_end=fg.bin_end,
                                                               ttype=fg.itransform_ttype),
                                        trim=trim))


def _power_uncompress(feature: AudioF) -> AudioF:
    import numpy as np

    mag = np.abs(feature)
    phase = np.angle(feature)
    mag = mag ** (1. / 0.3)
    real_uncompress = mag * np.cos(phase)
    imag_uncompress = mag * np.sin(phase)

    return real_uncompress + 1j * imag_uncompress



import os
from typing import Optional

from .types import TimeAlignedType


def load_text(audio: str | os.PathLike[str]) -> Optional[TimeAlignedType]:
    """Load time-aligned text data given a McGill-Speech audio file.

    :param audio: Path to the McGill-Speech audio file.
    :return: A TimeAlignedType object.
    """
    import string
    import struct

    from sonusai.mixture import get_sample_rate

    if not os.path.exists(audio):
        return None

    sample_rate = get_sample_rate(str(audio))

    with open(audio, mode='rb') as f:
        content = f.read()

    riff_id, file_size, wave_id = struct.unpack('<4si4s', content[:12])
    if riff_id.decode('utf-8') != 'RIFF':
        return None

    if wave_id.decode('utf-8') != 'WAVE':
        return None

    fmt_id, fmt_size = struct.unpack('<4si', content[12:20])

    if fmt_id.decode('utf-8') != 'fmt ':
        return None

    if fmt_size != 16:
        return None

    (_wave_format_tag,
     channels,
     _samples_per_sec,
     _avg_bytes_per_sec,
     _block_align,
     bits_per_sample) = struct.unpack('<hhiihh', content[20:36])

    i = 36
    samples = None
    text = None
    while i < file_size:
        chunk_id = struct.unpack('<4s', content[i:i + 4])[0].decode('utf-8')
        chunk_size = struct.unpack('<i', content[i + 4:i + 8])[0]

        if chunk_id == 'data':
            samples = chunk_size / channels / (bits_per_sample / 8)
            break

        if chunk_id == 'afsp':
            chunks = struct.unpack(f'<{chunk_size}s', content[i + 8:i + 8 + chunk_size])[0]
            chunks = chunks.decode('utf-8').split('\x00')
            for chunk in chunks:
                if chunk.startswith('text: "'):
                    text = chunk[7:-1].lower().translate(str.maketrans('', '', string.punctuation))
        i += 8 + chunk_size + chunk_size % 2

    if text and samples:
        return TimeAlignedType(start=0, end=samples / sample_rate, text=text)

    return None



from sonusai.mixture.datatypes import AudioStatsMetrics
from sonusai.mixture.datatypes import AudioT


def calc_audio_stats(audio: AudioT, win_len: float = None) -> AudioStatsMetrics:
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.mixture import Transformer

    args = ['stats']
    if win_len is not None:
        args.extend(['-w', str(win_len)])

    tfm = Transformer()

    _, _, out = tfm.build(input_array=audio,
                          sample_rate_in=SAMPLE_RATE,
                          output_filepath='-n',
                          extra_args=args,
                          return_output=True)

    stats = {}
    lines = out.split('\n')
    for line in lines:
        split_line = line.split()
        if len(split_line) == 0:
            continue
        value = split_line[-1]
        key = ' '.join(split_line[:-1])
        stats[key] = value

    return AudioStatsMetrics(
        dco=float(stats['DC offset']),
        min=float(stats['Min level']),
        max=float(stats['Max level']),
        pkdb=float(stats['Pk lev dB']),
        lrms=float(stats['RMS lev dB']),
        pkr=float(stats['RMS Pk dB']),
        tr=float(stats['RMS Tr dB']),
        cr=float(stats['Crest factor']),
        fl=float(stats['Flat factor']),
        pkc=int(stats['Pk count']),
    )



from dataclasses import dataclass
from typing import Callable
from typing import Optional

from sonusai.mixture import AudioT


@dataclass(frozen=True)
class ASRResult:
    text: str
    confidence: Optional[float] = None
    lang: Optional[str] = None
    lang_prob: Optional[float] = None
    duration: Optional[float] = None
    num_segments: Optional[int] = None
    asr_cpu_time: Optional[float] = None


def get_available_engines() -> list[str]:
    from importlib import import_module
    from pkgutil import iter_modules

    module = import_module('sonusai.utils.asr_functions')
    engines = [method for method in dir(module) if not method.startswith('_')]
    for _, name, _ in iter_modules():
        if name.startswith('sonusai_asr_'):
            module = import_module(f'{name}.asr_functions')
            for method in dir(module):
                if not method.startswith('_'):
                    engines.append(method)

    return engines


def _asr_fn(engine: str) -> Callable[..., ASRResult]:
    from importlib import import_module
    from pkgutil import iter_modules

    module = import_module('sonusai.utils.asr_functions')
    for method in dir(module):
        if method == engine:
            return getattr(module, method)

    for _, name, _ in iter_modules():
        if name.startswith('sonusai_asr_'):
            module = import_module(f'{name}.asr_functions')
            for method in dir(module):
                if method == engine:
                    return getattr(module, method)

    raise ValueError(f'engine {engine} not supported')


def calc_asr(audio: AudioT | str, engine: str, **config) -> ASRResult:
    """Run ASR on audio

    :param audio: Numpy array of audio samples or location of an audio file
    :param engine: ASR engine to use
    :param config: kwargs configuration parameters
    :return: ASRResult object containing text and confidence
    """
    from copy import copy

    import numpy as np

    from sonusai.mixture import read_audio

    if not isinstance(audio, np.ndarray):
        audio = copy(read_audio(audio))

    return _asr_fn(engine)(audio, **config)



"""sonusai onnx_predict

usage: onnx_predict [-hvlwr] [--include GLOB] [-i MIXID] MODEL DATA ...

options:
    -h, --help
    -v, --verbose               Be verbose.
    -i MIXID, --mixid MIXID     Mixture ID(s) to generate if input is a mixture database. [default: *].
    --include GLOB              Search only files whose base name matches GLOB. [default: *.{wav,flac}].
    -w, --write-wav             Calculate inverse transform of prediction and write .wav files

Run prediction (inference) using an ONNX model on a SonusAI mixture dataset or audio files from a glob path.
The ONNX Runtime (ort) inference engine is used to execute the inference.

Inputs:
    MODEL       ONNX model .onnx file of a trained model (weights are expected to be in the file).

    DATA        The input data must be one of the following:
                * WAV
                  Using the given model, generate feature data and run prediction. A model file must be
                  provided. The MIXID is ignored.

                * directory
                  Using the given SonusAI mixture database directory, generate feature and truth data if not found.
                  Run prediction. The MIXID is required.


Note there are multiple ways to process model prediction over multiple audio data files:
1. TSE (timestep single extension): mixture transform frames are fit into the timestep dimension and the model run as
   a single inference call.  If batch_size is > 1 then run multiple mixtures in one call with shorter mixtures
   zero-padded to the size of the largest mixture.
2. TME (timestep multi-extension): mixture is split into multiple timesteps, i.e. batch[0] is starting timesteps, ...
   Note that batches are run independently, thus sequential state from one set of timesteps to the next will not be
   maintained, thus results for such models (i.e. conv, LSTMs, in the timestep dimension) would not match using
   TSE mode.

TBD not sure below make sense, need to continue ??
2. BSE (batch single extension): mixture transform frames are fit into the batch dimension. This make sense only if
   independent predictions are made on each frame w/o considering previous frames (timesteps=1) or there is no
   timestep dimension in the model (timesteps=0).
3. Classification

Outputs the following to opredict-<TIMESTAMP> directory:
    <id>.h5
        dataset:    predict
    onnx_predict.log

"""
import signal


def signal_handler(_sig, _frame):
    import sys

    from sonusai import logger

    logger.info('Canceled due to keyboard interrupt')
    sys.exit(1)


signal.signal(signal.SIGINT, signal_handler)


def main() -> None:
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    wav = args['--write-wav']
    mixids = args['--mixid']
    include = args['--include']
    model_path = args['MODEL']
    data_paths = args['DATA']

    from os import makedirs
    from os.path import abspath
    from os.path import basename
    from os.path import isdir
    from os.path import join
    from os.path import normpath
    from os.path import realpath
    from os.path import splitext

    import h5py
    import numpy as np
    import onnxruntime as ort

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import get_audio_from_feature
    from sonusai.utils import PathInfo
    from sonusai.utils import braced_iglob
    from sonusai.utils import create_ts_name
    from sonusai.utils import load_ort_session
    from sonusai.utils import reshape_inputs
    from sonusai.utils import write_audio

    mixdb_path = None
    mixdb = None
    p_mixids = None
    entries: list[PathInfo] = []

    if len(data_paths) == 1 and isdir(data_paths[0]):
        # Assume it's a single path to SonusAI mixdb subdir
        in_basename = basename(normpath(data_paths[0]))
        mixdb_path = data_paths[0]
    else:
        # search all data paths for .wav, .flac (or whatever is specified in include)
        in_basename = ''

    output_dir = create_ts_name('opredict-' + in_basename)
    makedirs(output_dir, exist_ok=True)

    # Setup logging file
    create_file_handler(join(output_dir, 'onnx-predict.log'))
    update_console_handler(verbose)
    initial_log_messages('onnx_predict')

    providers = ort.get_available_providers()
    logger.info(f'Loaded ONNX Runtime, available providers: {providers}.')

    session, options, model_root, hparams, sess_inputs, sess_outputs = load_ort_session(model_path)
    if hparams is None:
        logger.error(f'Error: ONNX model does not have required SonusAI hyperparameters, cannot proceed.')
        raise SystemExit(1)
    if len(sess_inputs) != 1:
        logger.error(f'Error: ONNX model does not have 1 input, but {len(sess_inputs)}. Exit due to unknown input.')

    in0name = sess_inputs[0].name
    in0type = sess_inputs[0].type
    out_names = [n.name for n in session.get_outputs()]

    logger.info(f'Read and compiled ONNX model from {model_path}.')

    if mixdb_path is not None:
        # Assume it's a single path to SonusAI mixdb subdir
        logger.debug(f'Attempting to load mixture database from {mixdb_path}')
        mixdb = MixtureDatabase(mixdb_path)
        logger.info(f'SonusAI mixdb: found {mixdb.num_mixtures} mixtures with {mixdb.num_classes} classes')
        p_mixids = mixdb.mixids_to_list(mixids)
        if len(p_mixids) != mixdb.num_mixtures:
            logger.info(f'Processing a subset of {p_mixids} from available mixtures.')
    else:
        for p in data_paths:
            location = join(realpath(abspath(p)), '**', include)
            logger.debug(f'Processing {location}')
            for file in braced_iglob(pathname=location, recursive=True):
                name = file
                entries.append(PathInfo(abs_path=file, audio_filepath=name))
        logger.info(f'{len(data_paths)} data paths specified, found {len(entries)} audio files.')

    if in0type.find('float16') != -1:
        model_is_fp16 = True
        logger.info(f'Detected input of float16, converting all feature inputs to that type.')
    else:
        model_is_fp16 = False

    if mixdb_path is not None and hparams['batch_size'] == 1:
        # mixdb input
        # Assume (of course) that mixdb feature, etc. is what model expects
        if hparams['feature'] != mixdb.feature:
            logger.warning(f'Mixture feature does not match model feature, this inference run may fail.')
        # no choice, can't use hparams.feature since it's different from the mixdb
        feature_mode = mixdb.feature

        for mixid in p_mixids:
            # frames x stride x feature_params
            feature, _ = mixdb.mixture_ft(mixid)
            if hparams['timesteps'] == 0:
                # no timestep dimension, reshape will handle
                timesteps = 0
            else:
                # fit frames into timestep dimension (TSE mode)
                timesteps = feature.shape[0]

            feature, _ = reshape_inputs(feature=feature,
                                        batch_size=1,
                                        timesteps=timesteps,
                                        flatten=hparams['flatten'],
                                        add1ch=hparams['add1ch'])
            if model_is_fp16:
                feature = np.float16(feature)  # type: ignore
            # run inference, ort session wants i.e. batch x timesteps x feat_params, outputs numpy BxTxFP or BxFP
            predict = session.run(out_names, {in0name: feature})[0]
            # predict, _ = reshape_outputs(predict=predict[0], timesteps=frames)  # frames x feat_params
            output_fname = join(output_dir, mixdb.mixtures[mixid].name)
            with h5py.File(output_fname, 'a') as f:
                if 'predict' in f:
                    del f['predict']
                f.create_dataset('predict', data=predict)
            if wav:
                # note only makes sense if model is predicting audio, i.e., timestep dimension exists
                # predict_audio wants [frames, channels, feature_parameters] equivalent to timesteps, batch, bins
                predict = np.transpose(predict, [1, 0, 2])
                predict_audio = get_audio_from_feature(feature=predict, feature_mode=feature_mode)
                owav_name = splitext(output_fname)[0] + '_predict.wav'
                write_audio(owav_name, predict_audio)


if __name__ == '__main__':
    main()



import numpy as np

from sonusai.mixture.datatypes import Segsnr
from sonusai.mixture.datatypes import SnrFMetrics


def calc_snr_f(segsnr_f: Segsnr) -> SnrFMetrics:
    """Calculate metrics of snr_f truth data.

    For now, includes mean and variance of the raw values (usually energy)
    and mean and standard deviation of the dB values (10 * log10).
    """
    if np.count_nonzero(segsnr_f) == 0:
        # If all entries are zeros
        return SnrFMetrics(0, 0, -np.inf, 0)

    tmp = np.ma.array(segsnr_f, mask=np.logical_not(np.isfinite(segsnr_f)), dtype=np.float32)
    if np.ma.count_masked(tmp) == np.ma.size(tmp, axis=0):
        # If all entries are infinite
        return SnrFMetrics(np.inf, 0, np.inf, 0)

    snr_mean = np.mean(tmp, axis=0)
    snr_var = np.var(tmp, axis=0)

    tmp = 10 * np.ma.log10(tmp)
    if np.ma.count_masked(tmp) == np.ma.size(tmp, axis=0):
        # If all entries are masked, special case where all inputs are either 0 or infinite
        snr_db_mean = -np.inf
        snr_db_std = np.inf
    else:
        snr_db_mean = np.mean(tmp, axis=0)
        snr_db_std = np.std(tmp, axis=0)

    return SnrFMetrics(snr_mean, snr_var, snr_db_mean, snr_db_std)



import logging
from importlib import metadata
from os.path import dirname

__version__ = metadata.version(__package__)
BASEDIR = dirname(__file__)

commands_doc = """
   audiofe                      Audio front end
   calc_metric_spenh            Run speech enhancement and analysis
   doc                          Documentation
   genft                        Generate feature and truth data
   genmetrics                   Generate mixture metrics data
   genmix                       Generate mixture and truth data
   genmixdb                     Generate a mixture database
   gentcst                      Generate target configuration from a subdirectory tree
   lsdb                         List information about a mixture database
   mkwav                        Make WAV files from a mixture database
   onnx_predict                 Run ONNX predict on a trained model
   plot                         Plot mixture data
   post_spenh_targetf           Run post-processing for speech enhancement targetf data
   summarize_metric_spenh       Summarize speech enhancement and analysis results
   tplot                        Plot truth data
   vars                         List custom SonusAI variables
"""

# create logger
logger = logging.getLogger('sonusai')
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(message)s')
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)


class SonusAIError(Exception):
    def __init__(self, value):
        logger.error(value)


# create file handler
def create_file_handler(filename: str) -> None:
    fh = logging.FileHandler(filename=filename, mode='w')
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(formatter)
    logger.addHandler(fh)


# update console handler
def update_console_handler(verbose: bool) -> None:
    if not verbose:
        logger.removeHandler(console_handler)
        console_handler.setLevel(logging.INFO)
        logger.addHandler(console_handler)


# write initial log message
def initial_log_messages(name: str, subprocess: str = None) -> None:
    from datetime import datetime
    from getpass import getuser
    from os import getcwd
    from socket import gethostname
    from sys import argv

    if subprocess is None:
        logger.info(f'SonusAI {__version__}')
    else:
        logger.info(f'SonusAI {subprocess}')
    logger.info(f'{name}')
    logger.info('')
    logger.debug(f'Host:      {gethostname()}')
    logger.debug(f'User:      {getuser()}')
    logger.debug(f'Directory: {getcwd()}')
    logger.debug(f'Date:      {datetime.now()}')
    logger.debug(f'Command:   {" ".join(argv)}')
    logger.debug('')


def commands_list(doc: str = commands_doc) -> list[str]:
    lines = doc.split('\n')
    commands = []
    for line in lines:
        command = line.strip().split(' ').pop(0)
        if command:
            commands.append(command)
    return commands



from sonusai.mixture.datatypes import Truth
from sonusai.mixture.truth_functions.data import Data


def _core(data: Data, polar: bool) -> Truth:
    import numpy as np

    from sonusai import SonusAIError

    if data.config.num_classes != data.target_fft.bins:
        raise SonusAIError(f'Invalid num_classes for crm truth: {data.config.num_classes}')

    if data.target_fft.bins != data.noise_fft.bins:
        raise SonusAIError('Transform size mismatch for crm truth')

    for offset in data.offsets:
        target_f = data.target_fft.execute(data.target_audio[offset:offset + data.frame_size]).astype(np.complex64)
        noise_f = data.noise_fft.execute(data.noise_audio[offset:offset + data.frame_size]).astype(np.complex64)
        mixture_f = target_f + noise_f

        crm_data = np.empty(target_f.shape, dtype=np.complex64)
        with np.nditer(target_f, flags=['multi_index'], op_flags=[['readwrite']]) as it:
            for _ in it:
                num = target_f[it.multi_index]
                den = mixture_f[it.multi_index]
                if num == 0:
                    crm_data[it.multi_index] = 0
                elif den == 0:
                    crm_data[it.multi_index] = complex(np.inf, np.inf)
                else:
                    crm_data[it.multi_index] = num / den

        indices = slice(offset, offset + data.frame_size)

        def c1(c_data: np.ndarray, is_polar: bool) -> np.ndarray:
            if is_polar:
                return np.absolute(c_data)
            return np.real(c_data)

        def c2(c_data: np.ndarray, is_polar: bool) -> np.ndarray:
            if is_polar:
                return np.angle(c_data)
            return np.imag(c_data)

        for index in data.zero_based_indices:
            data.truth[indices, index:index + data.target_fft.bins] = c1(crm_data, polar)
            data.truth[indices, (index + data.target_fft.bins):(index + 2 * data.target_fft.bins)] = c2(crm_data, polar)

    return data.truth


def crm(data: Data) -> Truth:
    """Complex ratio mask truth generation function

Calculates the true complex ratio mask (CRM) truth which is a complex number
per bin = Mr + j*Mi. For a given noisy STFT bin value Y, it is used as

(Mr*Yr + Mi*Yi) / (Yr^2 + Yi^2) + j*(Mi*Yr - Mr*Yi)/ (Yr^2 + Yi^2)

Output shape: [:, bins]
    """
    return _core(data=data, polar=False)


def crmp(data: Data) -> Truth:
    """Complex ratio mask polar truth generation function

Same as the crm function except the results are magnitude and phase
instead of real and imaginary.

Output shape: [:, bins]
    """
    return _core(data=data, polar=True)
